import os.path
from dataclasses import dataclass
from typing import Dict, Optional

import yaml

from . import dtslogger
from .constants import DTShellConstants
from .exceptions import ConfigNotPresent, InvalidConfig


@dataclass
class ShellConfig:
    token_dt1: Optional[str]  ## key
    docker_username: Optional[str]
    docker_password: Optional[str]
    duckietown_version: Optional[str]  # daffy, master19, ...
    docker_credentials: Dict[str, Dict[str, str]]  # username, secret


@dataclass
class RepoInfo:
    username: str
    project: str
    branch: str


def RepoInfo_for_version(duckietown_version: str) -> RepoInfo:
    username = "duckietown"
    project = "duckietown-shell-commands"
    branch = duckietown_version
    return RepoInfo(username, project, branch)


def remoteurl_from_RepoInfo(ri: RepoInfo) -> str:
    return "https://github.com/%s/%s" % (ri.username, ri.project)


def get_shell_config_default() -> ShellConfig:
    return ShellConfig(
        token_dt1=None,
        docker_username=None,
        duckietown_version=None,
        docker_password=None,
        docker_credentials={},
    )


def get_config_path() -> str:
    root = DTShellConstants.ROOT
    config_path = os.path.expanduser(root)
    return config_path


def get_shell_config_file() -> str:
    config_path = get_config_path()
    config_file = os.path.join(config_path, "config.yaml")
    return config_file


def read_shell_config() -> ShellConfig:
    """Reads the config file. Raises InvalidConfig or ConfigNotPresent."""
    config_file = get_shell_config_file()
    return read_shell_config_from_file(config_file)


def write_shell_config(shell_config: ShellConfig) -> None:
    """Saves the she"""
    config_file = get_shell_config_file()
    write_shell_config_to_file(shell_config, config_file)


DT1_TOKEN_CONFIG_KEY = DTShellConstants.DT1_TOKEN_CONFIG_KEY
CONFIG_DOCKER_USERNAME = DTShellConstants.CONFIG_DOCKER_USERNAME
CONFIG_DOCKER_PASSWORD = DTShellConstants.CONFIG_DOCKER_PASSWORD
CONFIG_DUCKIETOWN_VERSION = DTShellConstants.CONFIG_DUCKIETOWN_VERSION
CONFIG_DOCKER_CREDENTIALS = DTShellConstants.CONFIG_DOCKER_CREDENTIALS


def write_shell_config_to_file(shell_config: ShellConfig, filename: str) -> None:
    data = {
        DT1_TOKEN_CONFIG_KEY: shell_config.token_dt1,
        CONFIG_DOCKER_USERNAME: shell_config.docker_username,
        CONFIG_DUCKIETOWN_VERSION: shell_config.duckietown_version,
        CONFIG_DOCKER_PASSWORD: shell_config.docker_password,
        CONFIG_DOCKER_CREDENTIALS: shell_config.docker_credentials,
    }
    dn = os.path.dirname(filename)
    if not os.path.exists(dn):
        os.makedirs(dn)
    with open(filename, "w") as f:
        s = yaml.dump(data)
        f.write(s)


def read_shell_config_from_file(fn: str) -> ShellConfig:
    """Raises InvalidConfig or ConfigNotPresent"""
    dtslogger.debug(f"reading config {fn}")

    if not os.path.exists(fn):
        raise ConfigNotPresent(fn)

    with open(fn, "r") as fp:
        data = fp.read()

    try:
        d = yaml.load(data, Loader=yaml.Loader)
    except BaseException as e:
        msg = f"Cannot read config file {fn}"
        raise InvalidConfig(msg) from e

    if d is None:
        raise ConfigNotPresent(fn)

    if not isinstance(d, dict):
        msg = f"Invalid config file {fn}; expected dict, got {type(d)}"
        raise InvalidConfig(msg)

    token_dt1 = d.pop(DT1_TOKEN_CONFIG_KEY, None)
    docker_username = d.pop(CONFIG_DOCKER_USERNAME, None)
    docker_password = d.pop(CONFIG_DOCKER_PASSWORD, None)
    duckietown_version = d.pop(CONFIG_DUCKIETOWN_VERSION, None)
    docker_credentials = d.pop(CONFIG_DOCKER_CREDENTIALS, {})
    if docker_credentials is None:
        docker_credentials = {}

    if "docker.io" not in docker_credentials and docker_username is not None:
        docker_credentials["docker.io"] = {"username": docker_username, "secret": docker_password}

    if d:
        msg = f"The config file {fn} contains other options that I do not understand: {d}"
        dtslogger.warning(msg)

    return ShellConfig(
        token_dt1=token_dt1,
        duckietown_version=duckietown_version,
        docker_username=docker_username,
        docker_password=docker_password,
        docker_credentials=docker_credentials,
    )



# -*- coding: utf-8 -*-
from typing import Optional

import termcolor

from . import __version__


class DTShellConstants:
    ROOT = "~/.dt-shell/"
    ENV_COMMANDS = "DTSHELL_COMMANDS"

    DT1_TOKEN_CONFIG_KEY = "token_dt1"
    CONFIG_DOCKER_USERNAME = "docker_username"
    CONFIG_DOCKER_PASSWORD = "docker_password"
    CONFIG_DUCKIETOWN_VERSION = "duckietown_version"
    CONFIG_DOCKER_CREDENTIALS = "docker_credentials"


ALLOWED_BRANCHES = ["ente(-[\w]+)?", "daffy(-[\w]+)?", "master19(-[\w]+)?", "devel(-[\w]+)?"]

DEBUG = False

CHECK_CMDS_UPDATE_EVERY_MINS = 5

DNAME = "Duckietown Shell"


def INTRO(extra: Optional[str] = None) -> str:
    return """

Welcome to the {Duckietown} ({version}).
{extra}
Type "help" or "?" to list commands.

""".format(
        Duckietown=termcolor.colored(DNAME, "yellow", attrs=["bold"]),
        version=__version__,
        extra=extra or ""
    ).lstrip()



__all__ = [
    "InvalidEnvironment",
    "UserError",
    "CommandsLoadingException",
    "ConfigNotPresent",
    "ConfigInvalid",
    "CouldNotGetVersion",
    "NoCacheAvailable",
    "URLException",
    "InvalidConfig",
]


class InvalidEnvironment(Exception):
    pass


class UserError(Exception):
    """an error that will be briefly printed"""

    pass


class InvalidConfig(Exception):
    pass


class CommandsLoadingException(Exception):
    pass


class ConfigInvalid(Exception):
    pass


class ConfigNotPresent(Exception):
    pass


class CouldNotGetVersion(Exception):
    pass


class URLException(Exception):
    pass


class NoCacheAvailable(Exception):
    pass



from typing import Optional, Sequence

import termcolor

from .utils import dark_yellow

__all__ = ["dts_print"]


def dts_print(msg: str, color: Optional[str] = None, attrs: Sequence[str] = ()) -> None:
    """
    Prints a message to the user.
    """
    msg = msg.strip()  # remove space
    print("")  # always separate
    lines = msg.split("\n")
    prefix = "dts : "
    filler = "    : "
    # filler = ' ' * len(prefix)

    for i, line in enumerate(lines):
        f = prefix if i == 0 else filler
        on_color = None
        line = termcolor.colored(line, color, on_color, list(attrs))
        s = "%s %s" % (dark_yellow(f), line)
        print(s)



import glob
import json
import os
import time
from os.path import getmtime
from typing import Dict, Optional

from . import dtslogger
from .config import remoteurl_from_RepoInfo, RepoInfo
from .constants import CHECK_CMDS_UPDATE_EVERY_MINS
from .logging import dts_print
from .utils import run_cmd


class InvalidRemote(Exception):
    pass


def _init_commands(commands_path: str, repo_info: RepoInfo) -> None:
    """Raises InvalidRemote if it cannot find it"""

    dtslogger.info("Downloading commands in %s ..." % commands_path)
    # the repo now exists
    remote_url = remoteurl_from_RepoInfo(repo_info)
    run_cmd(["git", "clone", "-b", repo_info.branch, "--recurse-submodules", remote_url, commands_path])


def check_commands_outdated(commands_path: str, repo_info: RepoInfo) -> bool:
    commands_update_check_flag = os.path.join(commands_path, ".updates-check")
    if not os.path.exists(commands_path):
        msg = "Repo does not exist."
        raise Exception(msg)

    dtslogger.info(f"looking at {commands_path}")
    local_sha = run_cmd(["git", "-C", commands_path, "rev-parse", "HEAD"])
    # get the first non-empty line
    local_sha = list(filter(len, local_sha.split("\n")))[0]
    # get remote SHA
    use_cached_sha = False
    if os.path.exists(commands_update_check_flag) and os.path.isfile(commands_update_check_flag):
        now = time.time()
        last_time_checked = getmtime(commands_update_check_flag)
        use_cached_sha = now - last_time_checked < CHECK_CMDS_UPDATE_EVERY_MINS * 60
    # get remote SHA
    if use_cached_sha:
        dtslogger.debug("Using cached remote SHA for command update check.")
        # no need to check now
        with open(commands_update_check_flag, "r") as fp:
            try:
                cached_check = json.load(fp)
            except ValueError:
                return False
            remote_sha = cached_check["remote"]
    else:
        dtslogger.debug("Fetching remote SHA for command update check from github.com")
        url = "https://api.github.com/repos/%s/%s/branches/%s" % (
            repo_info.username,
            repo_info.project,
            repo_info.branch,
        )
        try:
            from .version_check import get_url

            content = get_url(url)
            data = json.loads(content)
            remote_sha = data["commit"]["sha"]
        except BaseException as e:
            from .utils import format_exception

            dtslogger.error(format_exception(e))
            return False
    # check if we need to update
    need_update = local_sha != remote_sha
    if need_update:
        msg = """

An updated version of the commands is available.

Attempting auto-update.

        """
        dts_print(msg, color="yellow", attrs=["bold"])

        try:
            update_commands(commands_path, repo_info)

            # cache remote SHA
            if not use_cached_sha:
                save_update_check_flag(commands_path, remote_sha)

        except BaseException as e:
            from .utils import format_exception

            dtslogger.error(format_exception(e))

    # return success
    return True


def save_update_check_flag(commands_path: str, remote_sha: str) -> None:
    commands_update_check_flag = os.path.join(commands_path, ".updates-check")
    with open(commands_update_check_flag, "w") as fp:
        json.dump({"remote": remote_sha}, fp)


def update_commands(commands_path: str, repo_info: RepoInfo) -> bool:
    # create commands repo
    if not os.path.exists(commands_path):
        # the repo does not exist
        if not _init_commands(commands_path, repo_info):
            return False
    # the repo now exists
    dts_print("Updating commands...")
    # pull from origin (try 3 times)
    for trial in range(3):
        try:
            run_cmd(["git", "-C", commands_path, "pull", "--recurse-submodules", "origin", repo_info.branch])
        except RuntimeError as e:
            dtslogger.error(str(e))
            wait_time = 4
            th = {2: "nd", 3: "rd", 4: "th"}
            dtslogger.warning(
                "An error occurred while pulling the updated commands. Retrying for "
                f"the {trial + 2}-{th[trial + 2]} in {wait_time} seconds."
            )
            time.sleep(wait_time)
        else:
            break
    # update submodules
    run_cmd(["git", "-C", commands_path, "submodule", "update"])
    # get HEAD sha
    current_sha = run_cmd(["git", "-C", commands_path, "rev-parse", "HEAD"])
    # everything should be fine
    dts_print("OK")
    # cache current (local=remote) SHA
    save_update_check_flag(commands_path, current_sha)
    # return success
    return True


def _get_commands(path: str, lvl=0, all_commands=False) -> Optional[Dict[str, object]]:
    entries = glob.glob(os.path.join(path, "*"))
    files = [os.path.basename(e) for e in entries if os.path.isfile(e)]
    dirs = [e for e in entries if os.path.isdir(e) and (lvl > 0 or os.path.basename(e) != "lib")]
    # base case: empty dir
    if "command.py" not in files and not dirs:
        return None
    if (
        not all_commands
        and lvl == 1
        and ("installed.flag" not in files and "installed.user.flag" not in files)
    ):
        return None
    # check subcommands
    subcmds = {}
    for d in dirs:
        f = _get_commands(d, lvl + 1, all_commands)
        if f is not None:
            subcmds[os.path.basename(d)] = f
    # return
    if "command.py" not in files and not subcmds:
        return None
    # ---
    return subcmds



import locale
import logging
import os
import re
import sys
from typing import Dict, Union

import dt_shell
import yaml

from . import __version__, dtslogger
from .cli import DTShell, get_local_commands_info
from .cli_options import get_cli_options
from .config import get_shell_config_default, read_shell_config, write_shell_config
from .constants import ALLOWED_BRANCHES
from .env_checks import abort_if_running_with_sudo
from .exceptions import (
    CommandsLoadingException,
    ConfigInvalid,
    ConfigNotPresent,
    InvalidEnvironment,
    UserError,
)
from .logging import dts_print
from .utils import format_exception, href, replace_spaces
from .package_version_check import _get_installed_distributions


class OtherVersions:
    name2versions: Dict[str, Union[str, Dict[str, str]]] = {}


def cli_main() -> None:
    from .col_logging import setup_logging_color

    setup_logging_color()

    known_exceptions = (InvalidEnvironment, CommandsLoadingException)
    try:
        cli_main_()
    except UserError as e:

        msg = str(e)
        dts_print(msg, "red")
        print_version_info()
        sys.exit(1)
    except known_exceptions as e:
        msg = str(e)
        dts_print(msg, "red")
        print_version_info()
        sys.exit(1)
    except SystemExit:
        raise
    except KeyboardInterrupt:
        dts_print("User aborted operation.")
        pass
    except BaseException as e:
        msg = format_exception(e)
        dts_print(msg, "red", attrs=["bold"])
        print_version_info()
        sys.exit(2)


def print_version_info() -> None:
    v = OtherVersions.name2versions
    v["python"] = sys.version
    v["duckietown-shell"] = __version__
    try:
        shell_config = read_shell_config()
        commands_version = shell_config.duckietown_version
    except:
        commands_version = "ND"
    v["commands-version"] = commands_version

    v["encodings"] = {
        "stdout": sys.stdout.encoding,
        "stderr": sys.stderr.encoding,
        "locale": locale.getpreferredencoding(),
    }

    try:
        installed = _get_installed_distributions()
        pkgs = {_.project_name: _.version for _ in installed}
        for pkg_name, pkg_version in pkgs.items():
            include = (
                ("duckietown" in pkg_name)
                or ("dt-" in pkg_name)
                or ("-z" in pkg_name)
                or ("aido" in pkg_name)
            )
            if include:
                v[pkg_name] = pkg_version
    except ImportError:
        dtslogger.warning('Please update "pip" to have better debug info.')

    versions = yaml.dump(v, default_flow_style=False)
    # Please = termcolor.colored('Please', 'red', attrs=['bold'])
    fn = "~/shell-debug-info.txt"
    fn = os.path.expanduser(fn)
    with open(fn, "w") as f:
        f.write(versions)
    msg = f"""\
To report a bug, please also include the contents of {fn}
"""
    dts_print(msg, "red")


def print_info_command() -> None:
    url = href("https://github.com/duckietown/duckietown-shell-commands/issues")
    msg = """

    Problems with a command?

    Report here: {url}

    Troubleshooting:

    - If some commands update fail, delete ~/.dt-shell/commands

    - To reset the shell to "factory settings", delete ~/.dt-shell

      (Note: you will have to re-configure.)

        """.format(
        url=url
    )
    dts_print(msg)


def cli_main_() -> None:
    abort_if_running_with_sudo()

    # Problems with a step in the Duckiebot operation manual?
    #
    #     Report here: https://github.com/duckietown/docs-opmanual_duckiebot/issues

    # TODO: register handler for Ctrl-C
    cli_arguments = sys.argv[1:]
    cli_options, arguments = get_cli_options(cli_arguments)

    if not cli_options.quiet:
        print_info_command()

    # process options here
    if cli_options.debug:
        dtslogger.setLevel(logging.DEBUG)

    try:
        shell_config = read_shell_config()
    except ConfigInvalid as e:
        msg = "Cannot read the malformed config. Please delete the file."
        raise UserError(msg) from e
    except ConfigNotPresent:
        shell_config = get_shell_config_default()
        write_shell_config(shell_config)

    v = cli_options.set_version

    def is_allowed_branch(branch):
        allowed_braches_patterns = map(re.compile, ALLOWED_BRANCHES)
        for p in allowed_braches_patterns:
            if p.match(branch):
                return True
        return False

    if v is not None:
        if not is_allowed_branch(v):
            allowed_braches = [b.split("(")[0] for b in ALLOWED_BRANCHES]
            msg = f"Given version {v!r} is not one of {allowed_braches}."
            raise UserError(msg)
        shell_config.duckietown_version = v
        write_shell_config(shell_config)
        # need this so we can use in non-interactive settings
        sys.exit(0)

    if shell_config.duckietown_version is None:
        msg = """You have not specified a Duckietown version. Please use:

        dts --set-version <version>

        where <version> = daffy, master19
        """
        raise UserError(msg)
    if not cli_options.quiet:
        dtslogger.info(f"Commands version: {shell_config.duckietown_version}")
    commands_info = get_local_commands_info()
    # add subdirectory for version
    if not commands_info.leave_alone:
        commands_info.commands_path = os.path.join(
            commands_info.commands_path, shell_config.duckietown_version
        )

    shell = DTShell(shell_config, commands_info)

    # populate singleton
    dt_shell.shell = shell

    if arguments:
        arguments = map(replace_spaces, arguments)
        cmdline = " ".join(arguments)
        shell.onecmd(cmdline)
    else:
        shell.cmdloop()
